# Hyperscan helper worker setup
# Please don't modify this file as your changes might be overwritten with
# the next update.
#
# You can modify 'local.d/worker-hs_helper.inc' to add and merge
# parameters defined inside this section
#
# You can modify 'override.d/worker-hs_helper.inc' to strictly override all
# parameters defined inside this section
#
# See https://rspamd.com/doc/faq.html#what-are-the-locald-and-overrided-directories
# for details
# Directory to store compiled hyperscan databases
# cache_dir = "${DBDIR}";

# Cache backend: file (default), redis, http, or lua
# cache_backend = "file";

# Redis backend example:
# cache_backend = "redis";
# redis {
#   servers = "127.0.0.1:6379";
#   prefix = "rspamd_zhs";
#   ttl = 2592000;  # 30 days
#   compression = true;
# }

# Maximum time allowed for compiling a single expression (seconds)
# timeout = 1.0;

# Time between periodic recompilation checks (seconds)
# recompile = 60.0;
